/*
   +----------------------------------------------------------------------+
   | HipHop for PHP                                                       |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010-present Facebook, Inc. (http://www.facebook.com)  |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
*/
#include "hphp/hhbbc/hhbbc.h"

#include <stdexcept>
#include <iostream>
#include <cstdlib>
#include <string>
#include <memory>
#include <cstdint>
#include <algorithm>
#include <exception>
#include <utility>
#include <vector>
#include <filesystem>

#include <boost/program_options.hpp>

#include <folly/ScopeGuard.h>
#include <folly/String.h>
#include <folly/portability/Unistd.h>

#include "hphp/runtime/base/ini-setting.h"
#include "hphp/runtime/base/runtime-option.h"
#include "hphp/runtime/base/variable-serializer.h"
#include "hphp/runtime/base/vm-worker.h"
#include "hphp/hhvm/process-init.h"
#include "hphp/runtime/vm/native.h"
#include "hphp/runtime/vm/preclass-emitter.h"
#include "hphp/runtime/vm/repo-autoload-map-builder.h"
#include "hphp/runtime/vm/repo-file.h"
#include "hphp/runtime/vm/repo-global-data.h"
#include "hphp/runtime/vm/treadmill.h"

#include "hphp/hhbbc/misc.h"
#include "hphp/hhbbc/options.h"
#include "hphp/hhbbc/stats.h"
#include "hphp/hhbbc/parallel.h"
#include "hphp/hhbbc/representation.h"

#include "hphp/util/rds-local.h"
#include "hphp/util/logger.h"

namespace HPHP {

using namespace extern_worker;

namespace HHBBC {

namespace {

//////////////////////////////////////////////////////////////////////

std::string output_repo;
std::string input_repo;
bool logging = true;

//////////////////////////////////////////////////////////////////////

template<typename SinglePassReadableRange>
MethodMap make_method_map(SinglePassReadableRange& range) {
  auto ret = MethodMap{};
  for (auto& str : range) {
    std::vector<std::string> parts;
    folly::split("::", str, parts);
    if (parts.size() != 2) {
      ret[""].insert(str);
      continue;
    }
    ret[parts[0]].insert(parts[1]);
  }
  return ret;
}

template<typename SinglePassReadableRange>
OpcodeSet make_bytecode_map(SinglePassReadableRange& bcs) {
  if (bcs.empty()) return {};
  hphp_fast_map<std::string,Op> bcmap;
  for (auto i = 0; i < Op_count; i++) {
    auto const op = static_cast<Op>(i);
    bcmap[opcodeToName(op)] = op;
  }
  OpcodeSet oset;
  for (auto& n : bcs) {
    if (bcmap.count(n)) oset.insert(bcmap[n]);
  }
  return oset;
}

void parse_options(int argc, char** argv) {
  namespace po = boost::program_options;

  auto const defaultThreadCount =
    std::max<long>(sysconf(_SC_NPROCESSORS_ONLN) - 1, 1);
  auto const defaultFinalThreadCount =
    std::max<long>(defaultThreadCount - 1, 1);

  std::vector<std::string> trace_fns;
  std::vector<std::string> trace_bcs;
  bool no_logging = false;
  bool no_cores = false;

  po::options_description basic("Options");
  basic.add_options()
    ("help", "display help message")
    ("output,o",
      po::value(&output_repo)->default_value("hhvm.hhbbc"),
      "output hhbc repo path")
    ("input",
      po::value(&input_repo)->default_value("hhvm.hhbc"),
      "input hhbc repo path")
    ("stats-file",
      po::value(&options.stats_file)->default_value(""),
      "stats file path")
    ("no-optimizations",
      po::bool_switch(&options.NoOptimizations),
      "turn off all optimizations")
    ("no-logging",
      po::bool_switch(&no_logging),
      "turn off logging")
    ("no-cores",
      po::bool_switch(&no_cores),
      "turn off core dumps (useful when running lots of tests in parallel)")
    ("extended-stats",
      po::bool_switch(&options.extendedStats),
      "Spend time to produce extra stats")
    ("profile-memory",
     po::value(&options.profileMemory)->default_value(""),
      "If non-empty, dump jemalloc memory profiles at key points")
    ("parallel-num-threads",
      po::value(&parallel::num_threads)->default_value(defaultThreadCount),
      "Number of threads to use for parallelism")
    ("parallel-final-threads",
      po::value(&parallel::final_threads)->default_value(
        defaultFinalThreadCount),
      "Number of threads to use for the final pass")
    ("parallel-work-size",
      po::value(&parallel::work_chunk)->default_value(120),
      "Work unit size for parallelism")
    ("trace",
      po::value(&trace_fns)->composing(),
      "Add a function to increase tracing level on (for debugging)")
    ("trace-bytecode",
      po::value(&trace_bcs)->composing(),
      "Add a bytecode to trace (for debugging)")
    ;

  // Some extra esoteric options that aren't exposed in --help for
  // now.
  po::options_description extended("Extended Options");
  extended.add_options()
    ("analyze-func-wlimit",  po::value(&options.analyzeFuncWideningLimit))
    ("analyze-class-wlimit", po::value(&options.analyzeClassWideningLimit))
    ("return-refine-limit",  po::value(&options.returnTypeRefineLimit))
    ("public-sprop-refine-limit", po::value(&options.publicSPropRefineLimit))
    ("test-compression",     po::bool_switch(&options.TestCompression))
    ;

  po::options_description oflags("Optimization Flags");
  oflags.add_options()
    ("context-sensitive-interp",  po::value(&options.ContextSensitiveInterp))
    ("remove-dead-blocks",        po::value(&options.RemoveDeadBlocks))
    ("constant-prop",             po::value(&options.ConstantProp))
    ("constant-fold-builtins",    po::value(&options.ConstantFoldBuiltins))
    ("local-dce",                 po::value(&options.LocalDCE))
    ("global-dce",                po::value(&options.GlobalDCE))
    ("remove-unused-local-names", po::value(&options.RemoveUnusedLocalNames))
    ("compact-local-slots",       po::value(&options.CompactLocalSlots))
    ("insert-assertions",         po::value(&options.InsertAssertions))
    ("insert-stack-assertions",   po::value(&options.InsertStackAssertions))
    ("filter-assertions",         po::value(&options.FilterAssertions))
    ("strength-reduce",           po::value(&options.StrengthReduce))
    ("func-families",             po::value(&options.FuncFamilies))
    ("hard-private-prop",         po::value(&options.HardPrivatePropInference))
    ("analyze-public-statics",    po::value(&options.AnalyzePublicStatics))
    ;

  po::options_description eflags("Extern-Worker Flags");
  eflags.add_options()
    ("extern-worker-use-case",                 po::value(&options.ExternWorkerUseCase))
    ("extern-worker-working-dir",              po::value(&options.ExternWorkerWorkingDir))
    ("extern-worker-timeout-secs",             po::value(&options.ExternWorkerTimeoutSecs))
    ("extern-worker-throttle-retries",         po::value(&options.ExternWorkerThrottleRetries))
    ("extern-worker-throttle-base-wait-msecs", po::value(&options.ExternWorkerThrottleBaseWaitMSecs))
    ("extern-worker-force-subprocess",         po::value(&options.ExternWorkerForceSubprocess))
    ("extern-worker-use-exec-cache",           po::value(&options.ExternWorkerUseExecCache))
    ("extern-worker-cleanup",                  po::value(&options.ExternWorkerCleanup))
    ("extern-worker-use-rich-client",          po::value(&options.ExternWorkerUseRichClient))
    ("extern-worker-use-zippy-rich-client",    po::value(&options.ExternWorkerUseZippyRichClient))
    ("extern-worker-use-p2p",                  po::value(&options.ExternWorkerUseP2P))
    ("extern-worker-verbose-logging",          po::value(&options.ExternWorkerVerboseLogging))
    ("extern-worker-async-cleanup",            po::value(&options.ExternWorkerAsyncCleanup))
    ;

  po::options_description all;
  all.add(basic).add(extended).add(oflags).add(eflags);

  po::positional_options_description pd;
  pd.add("input", 1);

  po::variables_map vm;
  po::store(
    po::command_line_parser(argc, argv)
      .options(all)
      .positional(pd)
      .extra_parser(
        [&] (const std::string& s) -> std::pair<std::string,std::string> {
          if (s.find("-f") == 0) {
            if (s.substr(2, 3) == "no-") {
              return { s.substr(5), "false" };
            }
            return { s.substr(2), "true" };
          }
          return {};
        })
      .run(),
    vm
  );
  po::notify(vm);

  if (vm.count("help")) {
    std::cout << basic << "\n"
"Individual optimizations may be turned on and off using gcc-style -fflag\n"
"and -fno-flag arguments.  The various optimization flags are documented\n"
"in the code.\n";
    std::exit(0);
  }

  if (no_cores) {
    struct rlimit rl{};
    setrlimit(RLIMIT_CORE, &rl);
  }

  if (!options.ConstantProp) options.ConstantFoldBuiltins = false;

  options.TraceFunctions = make_method_map(trace_fns);
  options.TraceBytecodes = make_bytecode_map(trace_bcs);

  if (!options.profileMemory.empty()) {
    mallctlWrite("prof.active", true);
    mallctlWrite("prof.thread_active_init", true);
  }

  logging = !no_logging;
}

UNUSED void validate_options() {
  if (parallel::work_chunk <= 10 || parallel::num_threads < 1) {
    std::cerr << "Invalid parallelism configuration.\n";
    std::exit(1);
  }

  if (options.RemoveUnusedLocalNames && !options.GlobalDCE) {
    std::cerr << "-fremove-unused-local-names requires -fglobal-dce\n";
    std::exit(1);
  }

  if (options.CompactLocalSlots && !options.GlobalDCE) {
    std::cerr << "-fcompact-local-slots requires -fglobal-dce\n";
    std::exit(1);
  }
}

//////////////////////////////////////////////////////////////////////

RepoGlobalData get_global_data() {
  auto const now = std::chrono::high_resolution_clock::now();
  auto const nanos =
    std::chrono::duration_cast<std::chrono::nanoseconds>(
      now.time_since_epoch()
    );

  auto gd                        = RepoGlobalData{};
  gd.Signature                   = nanos.count();
  gd.HardGenericsUB              = RuntimeOption::EvalEnforceGenericsUB >= 2;
  gd.CheckPropTypeHints          = RuntimeOption::EvalCheckPropTypeHints;
  gd.HardPrivatePropInference    = options.HardPrivatePropInference;
  gd.PHP7_NoHexNumerics          = RuntimeOption::PHP7_NoHexNumerics;
  gd.PHP7_Substr                 = RuntimeOption::PHP7_Substr;
  gd.PHP7_Builtins               = RuntimeOption::PHP7_Builtins;
  gd.EnableIntrinsicsExtension   = RuntimeOption::EnableIntrinsicsExtension;
  gd.ForbidDynamicCallsToFunc    = RuntimeOption::EvalForbidDynamicCallsToFunc;
  gd.ForbidDynamicCallsToClsMeth =
    RuntimeOption::EvalForbidDynamicCallsToClsMeth;
  gd.ForbidDynamicCallsToInstMeth =
    RuntimeOption::EvalForbidDynamicCallsToInstMeth;
  gd.ForbidDynamicConstructs     = RuntimeOption::EvalForbidDynamicConstructs;
  gd.ForbidDynamicCallsWithAttr =
    RuntimeOption::EvalForbidDynamicCallsWithAttr;
  gd.LogKnownMethodsAsDynamicCalls =
    RuntimeOption::EvalLogKnownMethodsAsDynamicCalls;
  gd.AbortBuildOnVerifyError     = RuntimeOption::EvalAbortBuildOnVerifyError;
  gd.EnableArgsInBacktraces      = RuntimeOption::EnableArgsInBacktraces;
  gd.NoticeOnBuiltinDynamicCalls =
    RuntimeOption::EvalNoticeOnBuiltinDynamicCalls;
  gd.HackArrCompatSerializeNotices =
    RuntimeOption::EvalHackArrCompatSerializeNotices;
  gd.InitialNamedEntityTableSize  =
    RuntimeOption::EvalInitialNamedEntityTableSize;
  gd.InitialStaticStringTableSize =
    RuntimeOption::EvalInitialStaticStringTableSize;
  gd.EmitClassPointers = RuntimeOption::EvalEmitClassPointers;
  gd.EmitClsMethPointers = RuntimeOption::EvalEmitClsMethPointers;
  gd.IsVecNotices = RuntimeOption::EvalIsVecNotices;
  gd.RaiseClassConversionWarning =
    RuntimeOption::EvalRaiseClassConversionWarning;
  gd.ClassPassesClassname =
    RuntimeOption::EvalClassPassesClassname;
  gd.ClassnameNotices =
    RuntimeOption::EvalClassnameNotices;
  gd.ClassIsStringNotices = RuntimeOption::EvalClassIsStringNotices;
  gd.StrictArrayFillKeys = RuntimeOption::StrictArrayFillKeys;
  gd.TraitConstantInterfaceBehavior =
    RuntimeOption::EvalTraitConstantInterfaceBehavior;
  gd.BuildMayNoticeOnMethCallerHelperIsObject =
    RuntimeOption::EvalBuildMayNoticeOnMethCallerHelperIsObject;
  gd.DiamondTraitMethods = RuntimeOption::EvalDiamondTraitMethods;
  gd.EvalCoeffectEnforcementLevels = RO::EvalCoeffectEnforcementLevels;
  gd.SourceRootForFileBC = options.SourceRootForFileBC;

  for (auto const& elm : RuntimeOption::ConstantFunctions) {
    auto const s = internal_serialize(tvAsCVarRef(elm.second));
    gd.ConstantFunctions.emplace_back(elm.first, s.toCppString());
  }

  return gd;
}

//////////////////////////////////////////////////////////////////////

// Receives UnitEmitters and turns them into WholeProgramInput Key and
// Values. This "loads" them locally and uploads them into an
// extern_worker::Client in the way that whole_program expects.
struct LoadRepoJob {
  using W = WholeProgramInput;

  static std::string name() { return "hhbbc-load-repo"; }
  static void init(const Config& config) {
    process_init(config.o, config.gd, false);
  }
  static std::vector<W::Key> fini() {
    process_exit();
    return std::move(s_keys);
  }
  static Variadic<W::Value> run(UnitEmitterSerdeWrapper wrapper) {
    std::vector<W::Value> values;
    for (auto& [key, value] : W::make(std::move(wrapper.m_ue))) {
      s_keys.emplace_back(std::move(key));
      values.emplace_back(std::move(value));
    }
    return Variadic<W::Value>{std::move(values)};
  }

  static std::vector<W::Key> s_keys;
};
std::vector<WholeProgramInput::Key> LoadRepoJob::s_keys;
Job<LoadRepoJob> s_loadRepoJob;

// Load all UnitEmitters from the RepoFile and convert them into what
// whole_program expects as input.
WholeProgramInput load_repo(coro::TicketExecutor& executor,
                            Client& client) {
  trace_time timer("load repo");

  SCOPE_EXIT { RepoFile::destroy(); };
  RepoFile::loadGlobalTables(false);
  auto const units = RepoFile::enumerateUnits();
  if (logging) std::cout << folly::format("{} units\n", units.size());

  // Start this as early as possible
  coro::AsyncValue<Ref<Config>> config{
    [&client] () {
      return client.store(Config::get(RepoFile::globalData()));
    },
    executor.sticky()
  };

  // Shard the emitters using consistent hashing of their path.
  constexpr size_t kLoadGroupSize = 500;
  auto const numBuckets = (units.size() + kLoadGroupSize - 1) / kLoadGroupSize;

  std::vector<std::vector<const StringData*>> groups;
  groups.resize(numBuckets);

  for (auto const unit : units) {
    auto const idx = consistent_hash(unit->hash(), numBuckets);
    assertx(idx < numBuckets);
    groups[idx].emplace_back(unit);
  }

  // Maintain deterministic ordering within each group
  for (auto& group : groups) {
    std::sort(
      group.begin(),
      group.end(),
      [] (const StringData* a, const StringData* b) {
        return strcmp(a->data(), b->data()) < 0;
      }
    );
  }

  WholeProgramInput inputs;

  auto const load =
    [&] (std::vector<const StringData*> units) -> coro::Task<void> {

    // Load the UnitEmitters from disk (deferring this until here cuts
    // down on the number of UnitEmitters we have to keep in memory at
    // once).
    std::vector<UnitEmitterSerdeWrapper> ues;
    ues.reserve(units.size());
    for (auto const unit : units) {
      ues.emplace_back(
        RepoFile::loadUnitEmitter(unit, unit, Native::s_noNativeFuncs, false)
      );
    }

    // Store them (and the config we'll use).
    auto [refs, configRef] = HPHP_CORO_AWAIT(coro::collect(
      client.storeMulti(std::move(ues)),
      config.getCopy()
    ));

    std::vector<std::tuple<Ref<UnitEmitterSerdeWrapper>>> tuplized;
    tuplized.reserve(refs.size());
    for (auto& r : refs) tuplized.emplace_back(std::move(r));

    // Run the job and get refs to the keys and values.
    auto [valueRefs, keyRefs] = HPHP_CORO_AWAIT(
      client.exec(
        s_loadRepoJob,
        std::move(configRef),
        std::move(tuplized)
      )
    );

    // We need the keys locally, so load them. Values can stay as
    // Refs.
    auto keys = HPHP_CORO_AWAIT(client.load(std::move(keyRefs)));

    auto const numKeys = keys.size();
    size_t keyIdx = 0;
    for (auto& v : valueRefs) {
      for (auto& r : v) {
        always_assert(keyIdx < numKeys);
        inputs.add(std::move(keys[keyIdx]), std::move(r));
        ++keyIdx;
      }
    }
    always_assert(keyIdx == numKeys);

    HPHP_CORO_RETURN_VOID;
  };

  std::vector<coro::TaskWithExecutor<void>> tasks;
  for (auto& group : groups) {
    if (group.empty()) continue;
    tasks.emplace_back(load(std::move(group)).scheduleOn(executor.sticky()));
  }
  coro::wait(coro::collectRange(std::move(tasks)));
  return inputs;
}

extern_worker::Options make_extern_worker_options() {
  extern_worker::Options opts;
  opts
    .setUseCase(options.ExternWorkerUseCase)
    .setUseSubprocess(options.ExternWorkerForceSubprocess
                      ? extern_worker::Options::UseSubprocess::Always
                      : extern_worker::Options::UseSubprocess::Fallback)
    .setCacheExecs(options.ExternWorkerUseExecCache)
    .setCleanup(options.ExternWorkerCleanup)
    .setUseRichClient(options.ExternWorkerUseRichClient)
    .setUseZippyRichClient(options.ExternWorkerUseZippyRichClient)
    .setUseP2P(options.ExternWorkerUseP2P)
    .setVerboseLogging(options.ExternWorkerVerboseLogging);
  if (options.ExternWorkerTimeoutSecs > 0) {
    opts.setTimeout(std::chrono::seconds{options.ExternWorkerTimeoutSecs});
  }
  if (!options.ExternWorkerWorkingDir.empty()) {
    opts.setWorkingDir(options.ExternWorkerWorkingDir);
  }
  if (options.ExternWorkerThrottleRetries >= 0) {
    opts.setThrottleRetries(options.ExternWorkerThrottleRetries);
  }
  if (options.ExternWorkerThrottleBaseWaitMSecs >= 0) {
    opts.setThrottleBaseWait(
      std::chrono::milliseconds{options.ExternWorkerThrottleBaseWaitMSecs}
    );
  }
  return opts;
}

void compile_repo() {
  auto executor = std::make_unique<coro::TicketExecutor>(
    "HHBBCWorker",
    0,
    parallel::num_threads,
    [] {
      hphp_thread_init();
      hphp_session_init(Treadmill::SessionKind::HHBBC);
    },
    [] {
      hphp_context_exit();
      hphp_session_exit();
      hphp_thread_exit();
    },
    std::chrono::minutes{15}
  );
  auto client = std::make_unique<Client>(
    executor->sticky(),
    make_extern_worker_options()
  );

  auto inputs = load_repo(*executor, *client);

  RepoAutoloadMapBuilder autoload;
  RepoFileBuilder repo{output_repo};
  std::mutex repoLock;

  auto const emit = [&] (std::unique_ptr<UnitEmitter> ue) {
    autoload.addUnit(*ue);
    RepoFileBuilder::EncodedUE encoded{*ue};
    std::scoped_lock<std::mutex> _{repoLock};
    repo.add(encoded);
  };

  std::thread asyncDispose;
  SCOPE_EXIT { if (asyncDispose.joinable()) asyncDispose.join(); };
  auto const dispose = [&] (std::unique_ptr<coro::TicketExecutor> e,
                            std::unique_ptr<Client> c) {
    if (!options.ExternWorkerAsyncCleanup) {
      // If we don't want to cleanup asynchronously, do so now.
      c.reset();
      e.reset();
      return;
    }
    // All the thread does is reset the unique_ptr to run the dtor.
    asyncDispose = std::thread{
      [e = std::move(e), c = std::move(c)] () mutable {
        c.reset();
        e.reset();
      }
    };
  };

  HphpSession session{Treadmill::SessionKind::HHBBC};
  whole_program(
    std::move(inputs),
    std::move(executor),
    std::move(client),
    emit,
    dispose
  );

  trace_time timer{"finalizing repo"};
  repo.finish(get_global_data(), autoload);
}

//////////////////////////////////////////////////////////////////////

}

void process_init(const Options& o,
                  const RepoGlobalData& gd,
                  bool fullInit) {
  rds::local::init();
  SCOPE_FAIL { rds::local::fini(); };

  Hdf config;
  IniSetting::Map ini = IniSetting::Map::object;

  // We need to write correct coeffects before we load
  for (auto const& [name, value] : gd.EvalCoeffectEnforcementLevels) {
    config["Eval"]["CoeffectEnforcementLevels"][name] = value;
  }

  RO::Load(ini, config);
  RO::RepoAuthoritative                     = false;
  RO::EvalJit                               = false;
  RO::EvalLowStaticArrays                   = false;
  RO::RepoDebugInfo                         = false;
  Logger::LogLevel                          = Logger::LogError;

  // Load RepoGlobalData first because hphp_process_init can read
  // RuntimeOptions.
  gd.load(false);

  register_process_init(!fullInit);
  hphp_process_init(!fullInit);
  SCOPE_FAIL { hphp_process_exit(); };

  options = o;
  // Now that process state is set up, we can safely do a full load of
  // RepoGlobalData.
  gd.load();

  // When running hhbbc, these options are loaded from GD, and will
  // override CLI. When running hhvm, these options are not loaded
  // from GD, but read from CLI. NB: These are only needed if
  // RepoGlobalData::load() does not currently write them which is
  // called in hphp_process_init().
  RO::EvalForbidDynamicCallsToFunc        = gd.ForbidDynamicCallsToFunc;
  RO::EvalForbidDynamicCallsToClsMeth     = gd.ForbidDynamicCallsToClsMeth;
  RO::EvalForbidDynamicCallsToInstMeth    = gd.ForbidDynamicCallsToInstMeth;
  RO::EvalForbidDynamicConstructs         = gd.ForbidDynamicConstructs;
  RO::EvalForbidDynamicCallsWithAttr      = gd.ForbidDynamicCallsWithAttr;
  RO::EvalLogKnownMethodsAsDynamicCalls   = gd.LogKnownMethodsAsDynamicCalls;
  RO::EvalNoticeOnBuiltinDynamicCalls     = gd.NoticeOnBuiltinDynamicCalls;
  RO::EvalHackArrCompatSerializeNotices   = gd.HackArrCompatSerializeNotices;
  RO::EvalIsVecNotices                    = gd.IsVecNotices;
  RO::EvalEnforceGenericsUB               = gd.HardGenericsUB ? 2 : 1;

  options.SourceRootForFileBC             = gd.SourceRootForFileBC;
}

void process_exit() {
  hphp_process_exit();
  rds::local::fini();
}

int main(int argc, char** argv) try {
  parse_options(argc, argv);

  if (std::filesystem::exists(output_repo)) {
    std::cout << "output repo already exists; removing it\n";
    if (unlink(output_repo.c_str())) {
      std::cerr << "failed to unlink output repo: "
                << strerror(errno) << '\n';
      return 1;
    }
  }
  if (!std::filesystem::exists(input_repo)) {
    std::cerr << "input repo `" << input_repo << "' not found\n";
    return 1;
  }

  RepoFile::init(input_repo);

  auto const& gd = RepoFile::globalData();
  gd.load(false);
  if (gd.InitialNamedEntityTableSize) {
    RO::EvalInitialNamedEntityTableSize  = gd.InitialNamedEntityTableSize;
  }
  if (gd.InitialStaticStringTableSize) {
    RO::EvalInitialStaticStringTableSize = gd.InitialStaticStringTableSize;
  }

  process_init(options, gd, true);
  SCOPE_EXIT { process_exit(); };

  Logger::LogLevel = logging ? Logger::LogInfo : Logger::LogError;
  Logger::Escape = false;
  Logger::AlwaysEscapeLog = false;

  Trace::BumpRelease bumper(Trace::hhbbc_time, -1, logging);
  compile_repo();
  return 0;
}

catch (std::exception& e) {
  Logger::Error("std::exception: %s", e.what());
  return 1;
}

//////////////////////////////////////////////////////////////////////

}}
